/* Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA */

#include "dd/impl/types/spatial_reference_system_impl.h"

#include "gis/srs/srs.h"                   // gis::srs::parse_wkt
#include "dd/impl/dictionary_impl.h"       // Dictionary_impl
#include "dd/impl/sdi_impl.h"              // sdi read/write functions
#include "dd/impl/transaction_impl.h"      // Open_dictionary_tables_ctx
#include "dd/impl/raw/object_keys.h"       // id_key_type
#include "dd/impl/raw/raw_record.h"        // Raw_record
#include "dd/impl/tables/spatial_reference_systems.h" // Spatial_reference_sy...

using dd::tables::Spatial_reference_systems;

namespace dd {

///////////////////////////////////////////////////////////////////////////
// Spatial reference system implementation.
///////////////////////////////////////////////////////////////////////////

const Dictionary_object_table &Spatial_reference_system::OBJECT_TABLE()
{
  return Spatial_reference_systems::instance();
}

///////////////////////////////////////////////////////////////////////////

const Object_type &Spatial_reference_system::TYPE()
{
  static Spatial_reference_system_type s_instance;
  return s_instance;
}

///////////////////////////////////////////////////////////////////////////
// Spatial_reference_system_impl implementation.
///////////////////////////////////////////////////////////////////////////

bool Spatial_reference_system_impl::validate() const
{
  // The ID is an unsigned value, so we don't need to check the lower
  // bound.
  return id() > UINT32_MAX;
}

///////////////////////////////////////////////////////////////////////////

bool Spatial_reference_system_impl::restore_attributes(const Raw_record &r)
{
  bool error= false;

  restore_id(r, Spatial_reference_systems::FIELD_ID);
  restore_name(r, Spatial_reference_systems::FIELD_NAME);

  m_last_altered= r.read_int(Spatial_reference_systems::FIELD_LAST_ALTERED);
  m_created= r.read_int(Spatial_reference_systems::FIELD_CREATED);
  m_organization= r.read_str(Spatial_reference_systems::FIELD_ORGANIZATION);
  m_organization_coordsys_id= r.read_int(
                     Spatial_reference_systems::FIELD_ORGANIZATION_COORDSYS_ID);
  m_definition= r.read_str(Spatial_reference_systems::FIELD_DEFINITION);
  m_description= r.read_str(Spatial_reference_systems::FIELD_DESCRIPTION);

  gis::srs::Spatial_reference_system *srs= nullptr;
  // parse_wkt() will only allocate memory if successful.
  error= gis::srs::parse_wkt(id(), &m_definition, &srs);
  if (!error)
    m_parsed_definition.reset(srs);

  return error;
}

///////////////////////////////////////////////////////////////////////////

bool Spatial_reference_system_impl::store_attributes(Raw_record *r)
{
  Object_id default_catalog_id= Dictionary_impl::instance()->default_catalog_id();

  return store_id(r, Spatial_reference_systems::FIELD_ID) ||
         store_name(r, Spatial_reference_systems::FIELD_NAME) ||
         r->store(Spatial_reference_systems::FIELD_CATALOG_ID,
                  default_catalog_id) ||
         r->store(Spatial_reference_systems::FIELD_LAST_ALTERED,
                  m_last_altered) ||
         r->store(Spatial_reference_systems::FIELD_CREATED, m_created) ||
         r->store(Spatial_reference_systems::FIELD_ORGANIZATION,
                  m_organization) ||
         r->store(Spatial_reference_systems::FIELD_ORGANIZATION_COORDSYS_ID,
                  m_organization_coordsys_id) ||
         r->store(Spatial_reference_systems::FIELD_DEFINITION, m_definition) ||
         r->store(Spatial_reference_systems::FIELD_DESCRIPTION, m_description);
}

///////////////////////////////////////////////////////////////////////////

void Spatial_reference_system_impl::serialize(Sdi_wcontext *wctx, Sdi_writer *w)
  const
{
  w->StartObject();
  Entity_object_impl::serialize(wctx, w);
  write(w, m_last_altered, STRING_WITH_LEN("last_altered"));
  write(w, m_created, STRING_WITH_LEN("created"));
  write(w, m_organization, STRING_WITH_LEN("organization"));
  write(w, m_organization_coordsys_id,
        STRING_WITH_LEN("organization_coordsys_id"));
  write(w, m_definition, STRING_WITH_LEN("definition"));
  write(w, m_description, STRING_WITH_LEN("description"));
  w->EndObject();
}

///////////////////////////////////////////////////////////////////////////

bool Spatial_reference_system_impl::deserialize(Sdi_rcontext *rctx,
                                                const RJ_Value &val)
{
  Entity_object_impl::deserialize(rctx, val);
  read(&m_last_altered, val, "last_altered");
  read(&m_created, val, "created");
  read(&m_organization, val, "organization");
  read(&m_organization_coordsys_id, val, "organization_coordsys_id");
  read(&m_definition, val, "definition");
  read(&m_description, val, "description");

  gis::srs::Spatial_reference_system *srs= nullptr;
  // parse_wkt() will only allocate memory if successful.
  bool error= gis::srs::parse_wkt(id(), &m_definition, &srs);
  if (!error)
    m_parsed_definition.reset(srs);

  return error;
}

///////////////////////////////////////////////////////////////////////////

bool Spatial_reference_system::update_id_key(id_key_type *key, Object_id id)
{
  key->update(id);
  return false;
}

///////////////////////////////////////////////////////////////////////////

bool Spatial_reference_system::update_name_key(name_key_type *key,
                                               const std::string &name)
{
  return Spatial_reference_systems::update_object_key(key,
                      Dictionary_impl::instance()->default_catalog_id(),
                      name);
}

///////////////////////////////////////////////////////////////////////////
// Spatial_reference_system_type implementation.
///////////////////////////////////////////////////////////////////////////

void Spatial_reference_system_type::register_tables(
                      Open_dictionary_tables_ctx *otx) const
{
  otx->add_table<Spatial_reference_systems>();
}

///////////////////////////////////////////////////////////////////////////

}
